#ifndef SRPC_RPCCONTAINERS_H
#define SRPC_RPCCONTAINERS_H

#include "RpcTypes.h"
#ifdef _MSC_VER
#  pragma warning (push)
#  pragma warning (disable: 4702)
#endif
#include <list>
#include <map>
#include <set>
#include <vector>
#ifdef _MSC_VER
#  pragma warning (pop)
#endif
#include <algorithm>
#include <functional>
#include <cassert>

namespace srpc {

/** @addtogroup serialization
* @{
*/

struct NoReservePolicy
{
    template <typename Container>
    static inline void reserve(Container& /*container*/, size_t /*size*/) {}
};


struct VectorReservePolicy
{
    template <typename Container>
    static inline void reserve(Container& container, size_t size) {
        container.reserve(size);
    }
};

/// Standard C++ Library Sequence Container  RpcType
template <typename SeqContainer, typename ReservePolicy, size_t sizeBits>
class RpcContainer : public SeqContainer
{
public:
    typedef SeqContainer Base;
    typedef typename Base::value_type RpcType;
    typedef typename Base::iterator iterator;
public:
    void write(OStream& ostream) {
        ostream.write(static_cast<UInt32>(this->size()), sizeBits);
        iterator pos = this->begin();
        const iterator end = this->end();
        for (; pos != end; ++pos) {
            (*pos).write(ostream);
        }
    }

    void read(IStream& istream) {
        //assert(empty());
        this->clear();

        UInt32 size;
        istream.read(size, sizeBits);
        ReservePolicy::reserve(*this, size);
        for (UInt32 i = 0; i < size; ++i) {
            RpcType rt;
            rt.read(istream);
            this->push_back(rt);
        }
    }
};


/// std::vector  RpcType
template <typename RpcType, size_t sizeBits = Bits<UInt8>::size>
class RVector :
    public RpcContainer<std::vector<RpcType>, VectorReservePolicy, sizeBits>
{};


/// std::list  RpcType
template <typename RpcType, size_t sizeBits = Bits<UInt8>::size>
class RList :
    public RpcContainer<std::list<RpcType>, NoReservePolicy, sizeBits>
{};


/// std::set  RpcType
template <typename RpcType, size_t sizeBits = Bits<UInt8>::size>
class RSet : public std::set<RpcType>
{
public:
    typedef std::set<RpcType> Base;
    typedef typename Base::iterator iterator;
public:
    void write(OStream& ostream) {
        ostream.write(static_cast<UInt32>(this->size()), sizeBits);
        iterator pos = this->begin();
        const iterator end = this->end();
        for (; pos != end; ++pos) {
            const_cast<RpcType&>((*pos)).write(ostream);
        }
    }

    void read(IStream& istream) {
        //assert(empty());
        this->clear();

        UInt32 size;
        istream.read(size, sizeBits);
        for (UInt32 i = 0; i < size; ++i) {
            RpcType rt;
            rt.read(istream);
            this->insert(rt);
        }
    }
};


/// std::map  RpcType
template <typename K, typename V, size_t sizeBits = Bits<UInt8>::size>
class RMap : public std::map<K, V>
{
public:
    typedef std::map<K, V> Base;
    typedef typename Base::value_type value_type;
    typedef typename Base::iterator iterator;
public:
    void write(OStream& ostream) {
        ostream.write(static_cast<UInt32>(this->size()), sizeBits);
        iterator pos = this->begin();
        const iterator end = this->end();
        for (; pos != end; ++pos) {
            const_cast<K&>((*pos).first).write(ostream);
            const_cast<V&>((*pos).second).write(ostream);
        }
    }

    void read(IStream& istream) {
        //assert(empty());
        this->clear();

        UInt32 size;
        istream.read(size, sizeBits);
        for (UInt32 i = 0; i < size; ++i) {
            K key;
            key.read(istream);
            V value;
            value.read(istream);
            this->insert(value_type(key, value));
        }
    }
};

/** @} */ // addtogroup serialization

} // namespace srpc

#endif // SRPC_RPCCONTAINERS_H
